@@ -1,5 +1,5 @@
 /*
-    Copyright (C) 2001-2011 Ben Kibbey <bjk@luxsci.net>
+    Copyright (C) 2001-2015 Ben Kibbey <bjk@luxsci.net>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,6 +29,17 @@
 #include <pwd.h>
 #include <ctype.h>
 
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#ifndef LINE_MAX
+#ifdef _POSIX2_LINE_MAX
+#define LINE_MAX _POSIX2_LINE_MAX
+#else
+#define LINE_MAX 2048
+#endif
+#endif
+#endif
+
 #ifndef HAVE_STRSEP
 #include "../strsep.c"
 #endif
@@ -81,6 +92,9 @@ void ui_module_exit()
 
 static void free_logins(UTMP **u)
 {
+    if (!u)
+        return;
+
     if (login_count) {
 	UTMP **up;
 
@@ -99,7 +113,6 @@ static char *get_pid(uid_t uid, int multi)
     static int firstrun;
     static char line[LINE_MAX];
     int cnt, i;
-    pid_t pid = 0;
     char errbuf[LINE_MAX];
     struct kinfo_proc *kp;
 
@@ -111,9 +124,8 @@ static char *get_pid(uid_t uid, int multi)
     if (!kd) {
 	firstrun = 1;
 
-#ifdef __NetBSD__
-	if ((kd = kvm_openfiles(NULL, NULL, NULL,
-				O_RDONLY, errbuf)) == NULL) {
+#if defined(__NetBSD__) || defined (__OpenBSD__)
+	if ((kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf)) == NULL) {
 #else
 	if ((kd = kvm_openfiles(_PATH_DEVNULL, _PATH_DEVNULL, _PATH_DEVNULL,
 				O_RDONLY, errbuf)) == NULL) {
@@ -123,31 +135,46 @@ static char *get_pid(uid_t uid, int multi)
 	}
     }
 
+#ifdef __OpenBSD__
+    if ((kp = kvm_getprocs(kd, KERN_PROC_UID, uid, sizeof(struct kinfo_proc),
+	 &cnt)) == NULL) {
+	warnx("kvm_getprocs(): %s", kvm_geterr(kd));
+	return "!";
+    }
+#else
     if ((kp = kvm_getprocs(kd, KERN_PROC_UID, uid, &cnt)) == NULL) {
 	warnx("kvm_getprocs(): %s", kvm_geterr(kd));
 	return "!";
     }
+#endif
 
     for (i = 0; i < cnt; i++) {
 	char buf[32];
+	pid_t pid = 0;
 
+#ifdef __OpenBSD__
+	if ((kp[i].p_eflag & EPROC_SLEADER) && kp[i].p_tdev != -1)
+	    pid = kp[i].p_pid;
+#else
 #if __FreeBSD_version < 500000
-	if (kp[i].kp_eproc.e_flag & EPROC_SLEADER && kp[i].kp_eproc.e_tdev !=
-	    -1) {
+	if (kp[i].kp_eproc.e_flag & EPROC_SLEADER && kp[i].kp_eproc.e_tdev
+	    != -1) {
 	    pid = kp[i].kp_eproc.e_ppid;
 	    /*
 	     * pid = kp[i].kp_proc.p_pid;
 	     */
-
-	    if (pid == 1)
-		continue;
+	}
 #else
 	if (kp[i].ki_kiflag & KI_SLEADER && kp[i].ki_tdev != -1) {
 	    pid = kp[i].ki_pid;
-#endif
-	    snprintf(buf, sizeof(buf), "%i%c", pid, multi);
-	    safe_strncat(line, buf, sizeof(line));
 	}
+#endif
+#endif
+	if (!pid || pid == 1)
+	    continue;
+
+        snprintf(buf, sizeof(buf), "%i%c", pid, multi);
+	safe_strncat(line, buf, sizeof(line));
     }
 
     if (line[0] == '\0')
@@ -166,7 +193,7 @@ static char *get_pid(uid_t uid, int multi)
 #include <dirent.h>
 #endif
 
-#ifdef __svr4__
+#ifdef __sun__
 #include <unistd.h>
 #include <procfs.h>
 #endif
@@ -180,7 +207,7 @@ static char *get_pid(uid_t uid, int multi)
     pid_t *pids = 0, *tpids;
     int pid_index = 0;
 
-#ifdef __svr4__
+#ifdef __sun__
     int fd;
     struct pstatus pstat;
 #else
@@ -210,7 +237,7 @@ again:
 	char buf[LINE_MAX];
 	int i;
 
-#ifndef __svr4__
+#ifndef __sun__
 	char *t;
 #endif
 
@@ -227,10 +254,10 @@ again:
 	    continue;
 
 	/*
-	 * The current user owns this file (process id). 
+	 * The current user owns this file (process id).
 	 */
 	if (st.st_uid == uid) {
-#ifdef __svr4__
+#ifdef __sun__
 	    if ((fd = open(filename, O_RDONLY)) == -1)
 		continue;
 
@@ -267,7 +294,7 @@ again:
 #endif
 
 	    /*
-	     * Skip duplicate pids. 
+	     * Skip duplicate pids.
 	     */
 	    for (i = 0; i < pid_index; i++) {
 		if (pids[i] == pid)
@@ -357,6 +384,22 @@ static void last_strings(char *str)
 }
 
 /* Get the lastlog structure from the lastlog file. */
+#ifdef HAVE_GETLASTLOGX
+static char *lastlogin(const struct passwd *pw, char *tf)
+{
+    struct lastlogx *last =  getlastlogx(_PATH_LASTLOGX, pw->pw_uid, NULL);
+    static char buf[LINE_MAX];
+
+    if (!last)
+	return NULL;
+
+    snprintf(buf, sizeof(buf), "%s,%s,%s",
+	!last->ll_line[0] ? "!" : last->ll_line,
+	(!last->ll_host[0] || !isalnum(last->ll_host[0])) ?  "-" : last->ll_host,
+	stamp(last->ll_tv.tv_sec, tf));
+    return buf;
+}
+#else
 #if __FreeBSD_version >= 900000
 static char *lastlogin(const struct passwd *pw, char *tf)
 {
@@ -428,12 +471,13 @@ static char *lastlogin(const struct passwd *pw, char *tf)
 
     snprintf(buf, sizeof(buf), "%s,%s,%s",
 	!last.ll_line[0] ? "!" : last.ll_line,
-	(!last.ll_host[0] || !isalnum(last.ll_host[0])) ? 
+	(!last.ll_host[0] || !isalnum(last.ll_host[0])) ?
 	    !isdigit(last.ll_line[3]) ? "!" : "-" : last.ll_host,
 	!last.ll_time ? "!" : stamp(last.ll_time, tf));
     return buf;
 }
 #endif
+#endif
 
 /* This will return an array of utmp structures if a user is logged in, NULL
  * otherwise. We'll try to keep the utmp file descriptor open if possible to
@@ -441,7 +485,7 @@ static char *lastlogin(const struct passwd *pw, char *tf)
 static UTMP **get_utmp(const char *user)
 {
     UTMP **logins = NULL;
-#ifdef HAVE_UTMPX_H
+#if defined HAVE_UTMPX_H && defined (HAVE_SETUTXENT)
     UTMP *u;
 #else
     UTMP u;
@@ -461,7 +505,7 @@ static UTMP **get_utmp(const char *user)
 
     login_count = 0;
 
-#ifdef HAVE_UTMPX_H
+#if defined HAVE_UTMPX_H && defined (HAVE_SETUTXENT)
     setutxent();
 
     while ((u = getutxent()) != NULL) {
@@ -490,7 +534,7 @@ static UTMP **get_utmp(const char *user)
 		return NULL;
 	    }
 
-#ifdef HAVE_UTMPX_H
+#if defined HAVE_UTMPX_H && defined (HAVE_SETUTXENT)
 #ifdef __NetBSD__
 	    memcpy(logins[login_count]->ut_name, u->ut_name, UTX_NAMESIZE);
 	    logins[login_count]->ut_name[UTX_NAMESIZE-1] = 0;
@@ -688,7 +732,7 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 		add_string(&strings, line);
 		break;
 	    case 'm':
-		add_string(&strings, msgstat(u, multi));
+		add_string(&strings, u ? msgstat(u, multi) : "!");
 		break;
 	    case 't':
 		for (i = 0; i < login_count; i++) {
