@@ -1,5 +1,5 @@
 /*
-    Copyright (C) 2001-2011 Ben Kibbey <bjk@luxsci.net>
+    Copyright (C) 2001-2015 Ben Kibbey <bjk@luxsci.net>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -29,6 +29,17 @@
 #include <pwd.h>
 #include <time.h>
 
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#ifndef LINE_MAX
+#ifdef _POSIX2_LINE_MAX
+#define LINE_MAX _POSIX2_LINE_MAX
+#else
+#define LINE_MAX 2048
+#endif
+#endif
+#endif
+
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #endif
@@ -62,7 +73,7 @@ char *stamp(time_t epoch, const char *format)
     return buf;
 }
 
-/* 
+/*
  * This may be used in modules to add a string to the buffer (ui_module_exec()).
  */
 void add_string(char ***buf, const char *str)
@@ -168,11 +179,13 @@ static int open_module(char *filename, struct module_s **ptr)
     int dup = 0;
 
     strncpy(s, filename, sizeof(s));
+    s[sizeof(s)-1] = 0;
 
     if ((p = strrchr(s, '/')) != NULL)
 	p++;
     else {
 	strncpy(s, filename, sizeof(s));
+        s[sizeof(s)-1] = 0;
 	p = s;
     }
 
@@ -197,6 +210,7 @@ static int open_module(char *filename, struct module_s **ptr)
     modules[module_total] = NULL;
     mod->m = m;
     strncpy(mod->name, p, sizeof(mod->name));
+    mod->name[sizeof(mod->name)-1] = 0;
 
     if ((init = dlsym(mod->m, "ui_module_init")) == NULL)
 	warnx("%s", dlerror());
@@ -531,6 +545,7 @@ static char *get_home_directory()
     }
 
     strncpy(dir, pw->pw_dir, sizeof(dir));
+    dir[sizeof(dir)-1] = 0;
     return dir;
 }
 
@@ -612,9 +627,11 @@ static int parse_rc_file(const char *filename)
 	if (*p == '~') {
 	    s = get_home_directory();
 	    strncpy(tmp, s, sizeof(tmp));
+            tmp[sizeof(tmp)-1] = 0;
 	    p++;
 	    safe_strncat(tmp, p, sizeof(tmp));
 	    strncpy(name, tmp, sizeof(name));
+            name[sizeof(name)-1] = 0;
 	}
 
 	if (open_module(name, &mod))
